import 'dart:math';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'dart:ui' as ui;
import 'dart:typed_data';
import 'math_more.dart';
num degToRad(num deg) => deg * (pi / 180.0);

num radToDeg(num rad) => rad * (180.0 / pi);
//double get percent => (progress - min).toDouble() / (max - min);
int toProgress(double percent, {int min = 0, int max = 100}) {
  return (percent * (max - min)).toInt() + min;
}
double calScale(double targetLength, double originLength,double ratio){
  return targetLength * ratio / originLength;
}
Future<ui.Image> load(String asset) async {
  if(asset == null || asset == "") return null;
  ByteData data = await rootBundle.load(asset);
  ui.Codec codec = await ui.instantiateImageCodec(data.buffer.asUint8List());
  ui.FrameInfo fi = await codec.getNextFrame();
  return fi.image;
}

typedef OnProgressChanged<int> = void Function(int value);

class Knob extends StatefulWidget {
  final double radius;
  final double dotRadius;
  final double shadowWidth;
  final Color shadowColor;
  final Color dotColor;
  final Color dotEdgeColor;
  final Color ringColor;
  final OnProgressChanged onProgressChanged;
  final int progress;
  final int min,max;
  final int startAngleOffset;
  final String mainImagePath,backImagePath;
  final double mainRatio,backRatio;
  final bool touchAble;
  const Knob({
    Key key,
    @required this.radius,
    @required this.dotRadius,
    @required this.dotColor,
    this.shadowWidth = 2.0,
    this.shadowColor = Colors.black12,
    this.ringColor = const Color(0XFFF7F7FC),
    this.dotEdgeColor = const Color(0XFFF5F5FA),
    this.progress,
    this.onProgressChanged,
    this.min = 0,
    this.max = 100,
    this.startAngleOffset = 30,
    this.mainImagePath = 'assets/images/knob_main_n.png',
    this.backImagePath = 'assets/images/knob_back_n.png',
    this.mainRatio = -1, this.backRatio = -1,
    this.touchAble = true
  }) : super(key: key);

  @override
  State<StatefulWidget> createState() => _CircleProgressState();
}

class _CircleProgressState extends State<Knob> with SingleTickerProviderStateMixin {
  AnimationController progressController;
  final GlobalKey paintKey = GlobalKey();
  double get _width {
    RenderBox box = paintKey.currentContext.findRenderObject();
    return  box.size.width;
  }
  double get _height {
    RenderBox box = paintKey.currentContext.findRenderObject();
    return  box.size.height;
  }
  double currTouchDeg = 0, deg = 0, downTouchDeg = 0;

  final isAntiClockwise = false;
  bool _prepare = false;
  ui.Image _mainImage;
  ui.Image _backImage;
  @override
  void initState() {
    super.initState();
    loadAsset() async{
      num start = DateTime.now().millisecondsSinceEpoch;
      _mainImage = await load(widget.mainImagePath);
      _backImage = await load(widget.backImagePath);
      setState(() {
        _prepare = true;
      });
    }
    loadAsset();
    progressController = AnimationController(duration: Duration(milliseconds: 300), vsync: this);
    if (widget.progress != null) {
      var percent = (widget.progress - widget.min).toDouble() / (widget.max - widget.min);
      progressController.value = percent;
      deg = (progressController.value * (360 - widget.startAngleOffset * 2));
    }

    progressController.addListener(() {
      if (widget.onProgressChanged != null) {
        final progress = toProgress(
          progressController.value,
          min: widget.min,
          max: widget.max,
        );
        widget.onProgressChanged(progress);
      }
      setState(() {});
    });
  }

  @override
  void dispose() {
    progressController.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final double width = widget.radius * 2.0;
    final size = new Size(width, width);
    final angle = (progressController.value * (360 - widget.startAngleOffset * 2)).floor();
    return GestureDetector(
      onPanDown: _onPanDown,
      onPanStart: _onPanStart,
      onPanUpdate: _onPanUpdate,
      onPanEnd: _onPanEnd,
      child: Container(
        alignment: FractionalOffset.center,
        child: CustomPaint(
          key: paintKey,
          size: size,
          painter: _prepare ? ProgressPainter(
            dotRadius: widget.dotRadius,
            shadowWidth: widget.shadowWidth,
            shadowColor: widget.shadowColor,
            ringColor: widget.ringColor,
            dotColor: widget.dotColor,
            dotEdgeColor: widget.dotEdgeColor,
            angle: angle.toDouble(),
            startAngleOffset: widget.startAngleOffset,
            mainImage: _mainImage,
            backImage: _backImage,
            mainRatio: widget.mainRatio,
            backRatio: widget.backRatio,
          ) : null,
        ),
      ),
    );
  }
  void _onPanDown(DragDownDetails details){
      RenderBox getBox = paintKey.currentContext.findRenderObject();
      Offset local = getBox.globalToLocal(details.globalPosition);
      double dx = local.dx - _width / 2;
      double dy = local.dy - _height / 2;
      downTouchDeg = ((atan2(dy, dx) * 180) / pi);
  }
  void _onPanStart(DragStartDetails details) {
    if (!widget.touchAble) {
      return;
    }
    RenderBox getBox = paintKey.currentContext.findRenderObject();
    Offset local = getBox.globalToLocal(details.globalPosition);
  }

  void _onPanUpdate(DragUpdateDetails details) {
    if (!widget.touchAble) {
      return;
    }

    RenderBox getBox = paintKey.currentContext.findRenderObject();
    Offset local = getBox.globalToLocal(details.globalPosition);
    final double center = widget.radius;
    double dx = local.dx - _width / 2;
    double dy = local.dy - _height / 2;
    currTouchDeg = (atan2(dy, dx) * 180) / pi;
    double ddeg = currTouchDeg - downTouchDeg;
    if(abs(ddeg) > 90){
    }else{
      setDeg(deg + ddeg);
    }
    downTouchDeg = currTouchDeg;
  }
  void setDeg(double newDeg) {
    double maxDeg = (360 - widget.startAngleOffset * 2).toDouble();
    deg = newDeg;
    if(newDeg < 0){
      deg = 0;
    }else if (deg > maxDeg){
      deg = maxDeg;
    }
    progressController.value =  newDeg / (360 - widget.startAngleOffset * 2);
  }
  void _onPanEnd(DragEndDetails details) {
    if (!widget.touchAble) {
      return;
    }
  }

  /*bool _checkValidTouch(Offset pointer) {
    final double validInnerRadius = widget.radius - widget.dotRadius * 3;
    final double dx = pointer.dx;
    final double dy = pointer.dy;
    final double distanceToCenter = sqrt(pow(dx - widget.radius, 2) + pow(dy - widget.radius, 2));
    if (distanceToCenter < validInnerRadius ||
        distanceToCenter > widget.radius) {
      return false;
    }
    return true;
  }*/
}

class ProgressPainter extends CustomPainter {
  final double dotRadius;
  final double shadowWidth;
  final Color shadowColor;
  final Color dotColor;
  final Color dotEdgeColor;
  final Color ringColor;
  final double angle;
  final int startAngleOffset;
  final ui.Image mainImage;
  final ui.Image backImage;
  final double mainRatio,backRatio;
  ProgressPainter({
    this.dotRadius,
    this.shadowWidth = 2.0,
    this.shadowColor = Colors.black12,
    this.ringColor = const Color(0XFFF7F7FC),
    this.dotColor,
    this.dotEdgeColor = const Color(0XFFF5F5FA),
    this.angle = 0,
    this.startAngleOffset = 30,
    this.mainImage,
    this.backImage,
    this.mainRatio = -1, this.backRatio = -1,
  });

  @override
  void paint(Canvas canvas, Size size) {
    final double centerW = size.width * 0.5;
    final double centerH = size.height * 0.5;
    final center = centerW < centerH ? centerW : centerH;
    final Offset offsetCenter = Offset(centerW, centerH);
    final double drawRadius = size.width * 0.5 - dotRadius;
    final double radians = degToRad(angle);

    final double radiusOffset = dotRadius * 0.4;
    final double outerRadius = center;
    final double innerRadius = center - dotRadius * 2 + radiusOffset;

    final Rect arcRect = Rect.fromCircle(center: offsetCenter, radius: drawRadius);
    // draw shadow.
    final shadowPaint = Paint()
      ..style = PaintingStyle.stroke
      ..color = shadowColor
      ..strokeWidth = shadowWidth
      ..maskFilter = MaskFilter.blur(BlurStyle.normal, shadowWidth);
    //    canvas.drawCircle(offsetCenter, outerRadius, shadowPaint);
    //    canvas.drawCircle(offsetCenter, innerRadius, shadowPaint);
    Path path = Path.combine(
        PathOperation.difference,
        Path()
          ..addOval(Rect.fromCircle(center: offsetCenter, radius: outerRadius)),
        Path()
          ..addOval(
              Rect.fromCircle(center: offsetCenter, radius: innerRadius)));
//    canvas.drawShadow(path, shadowColor, 4.0, true);
    final paint = Paint()
      ..strokeWidth = 1
      ..isAntiAlias = true;

    // draw ring.
    final ringPaint = Paint()
      ..style = PaintingStyle.stroke
      ..color = ringColor
      ..strokeCap = StrokeCap.round
      ..strokeWidth = (outerRadius - innerRadius);

    canvas.save();
    if(backImage != null){
      double backScale = backRatio == -1 ? 1 : calScale(center * 2, backImage.width.toDouble(), backRatio);
      Matrix4 backMatrix = Matrix4.translationValues(centerW, centerH, 0)
        ..scale(backScale);
      canvas.transform(backMatrix.storage);
      canvas.drawImage(backImage, Offset(-backImage.width / 2, -backImage.height / 2), paint);
      canvas.restore();
      canvas.save();
    }
    double mainScale = mainRatio == -1 ? 1 : calScale(center * 2, mainImage.width.toDouble(), mainRatio);
    Matrix4 mainMatrix = Matrix4.translationValues(centerW, centerH, 0)
      ..scale(mainScale)
      ..rotateZ(degToRad(angle + 180 + startAngleOffset).toDouble());
    canvas.transform(mainMatrix.storage);
    canvas.drawImage(mainImage,Offset(-mainImage.width / 2, -mainImage.height / 2), paint);
    canvas.restore();
    canvas.save();
    canvas.drawArc(arcRect, degToRad(startAngleOffset + 90), degToRad(360 - startAngleOffset * 2), false, ringPaint);

    final Color currentDotColor = Color.alphaBlend(
        dotColor.withOpacity(0.7 + (0.3 * angle / 360)), Colors.white);

    // draw .
    if (angle > 0.0) {
      final progressWidth = outerRadius - innerRadius + radiusOffset;
      canvas.translate(size.width,0);
      canvas.rotate(degToRad(90));
      final Gradient gradient = new SweepGradient(
        endAngle: degToRad(360),
        colors: [
          Colors.white,
          currentDotColor,
        ],
      );
      final progressPaint = Paint()
        ..style = PaintingStyle.stroke
        ..strokeCap = StrokeCap.round
        ..strokeWidth = progressWidth
        ..shader = gradient.createShader(arcRect);
      canvas.drawArc(arcRect, degToRad(startAngleOffset), degToRad(angle), false, progressPaint);
    }
    canvas.restore();
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return true;
  }
}
